# src\file_conversor\dependency\abstract_pkg_manager.py

"""
This module provides functionalities for handling external backends.
"""

import platform
import shutil
import subprocess

from rich import print
from typing import Any, Callable, Iterable

# user-provided imports
from file_conversor.config import Log
from file_conversor.config.locale import get_translation

LOG = Log.get_instance()

_ = get_translation()
logger = LOG.getLogger(__name__)


class AbstractPackageManager:

    def __init__(self, dependencies: dict[str, str]) -> None:
        """
        Initializes package manager.

        :param dependencies: External dependencies to check. Format {``executable: dependency``}.        
        """
        super().__init__()
        self._os_type = platform.system()
        self._dependencies = dependencies
        self._pre_install_dep_callbacks: list[Callable[[], Any]] = []
        self._post_install_dep_callbacks: list[Callable[[], Any]] = []

    def check_dependencies(self) -> set[str]:
        """
        Check if external dependencies are installed.

        :return: Missing dependencies list.
        """
        missing_dependencies = set([])

        # check if executable exists
        for executable, dependency in self._dependencies.items():
            found_exe = shutil.which(executable)
            if not found_exe:
                logger.warning(f"ABSTRACT PKG MANAGER - {_('Executable')} '{executable}' {_('not found')}. {_('Marking dependency')} '{dependency}' {_('for installation')}.")
                missing_dependencies.add(dependency)

        # missing dependencies
        return missing_dependencies

    def get_pkg_manager_installed(self) -> str | None:
        """
        Checks if the package manager is already installed in system.

        :return: package manager path if installed, otherwise None

        :raises RuntimeError: if package manager not supported in the OS.
        """
        if self._os_type not in self.get_supported_oses():
            raise RuntimeError(f"{_("Package manager is not supported on")} '{self._os_type}'.")
        return self._get_pkg_manager_installed()

    def get_supported_oses(self) -> set[str]:
        """
        Gets the package manager supported OSes.

        :return: Set([PLATFORM_WINDOWS, PLATFORM_LINUX, PLATFORM_MACOS, ...])
        """
        return self._get_supported_oses()

    def install_pkg_manager(self) -> str | None:
        """
        Installs package manager for the current user.

        :return: package manager path if installation success, otherwise None .

        :raises RuntimeError: if package manager not supported in the OS.
        :raises RuntimeError: package manager already installed in system.
        :raises subprocess.CalledProcessError: package manager could not be installed in system.
        """
        if self.get_pkg_manager_installed():
            raise RuntimeError("Package manager already installed in system.")
        logger.info(f"{_('Installing package manager')} ...")
        logger.debug(f"{self._get_cmd_install_pkg_manager()}")
        subprocess.run(self._get_cmd_install_pkg_manager(), check=True)
        self._post_install_pkg_manager()
        return self.get_pkg_manager_installed()

    def install_dependencies(self, dependencies: Iterable[str]):
        """
        Installs dependencies with package manager.

        :param dependencies: External dependencies to install. Format [``dependency``, ...].

        :raises RuntimeError: package manager NOT installed in system.
        :raises subprocess.CalledProcessError: dependency could not be installed in system.
        """
        if not self.get_pkg_manager_installed():
            raise RuntimeError("Package manager NOT installed in system.")
        logger.info(f"{_('Running pre-install dependency commands')} ...")
        for callback in self._pre_install_dep_callbacks:
            callback()
        logger.info(f"{_('Installing missing dependencies')} {list(dependencies)} ...")
        for dep in dependencies:
            subprocess.run(self._get_cmd_install_dep(dep), check=True)
        logger.info(f"{_('Running post-install dependency commands')} ...")
        for callback in self._post_install_dep_callbacks:
            callback()

    def add_pre_install_callback(self, callback: Callable[[], Any]):
        """ Adds a pre-install callback to run before installing dependencies. """
        self._pre_install_dep_callbacks.append(callback)

    def add_post_install_callback(self, callback: Callable[[], Any]):
        """ Adds a post-install callback to run after installing dependencies. """
        self._post_install_dep_callbacks.append(callback)

    def _get_pkg_manager_installed(self) -> str | None:
        raise NotImplementedError("Method not overloaded.")

    def _get_supported_oses(self) -> set[str]:
        raise NotImplementedError("Method not overloaded.")

    def _get_cmd_install_pkg_manager(self) -> list[str]:
        raise NotImplementedError("Method not overloaded.")

    def _post_install_pkg_manager(self) -> None:
        raise NotImplementedError("Method not overloaded.")

    def _get_cmd_install_dep(self, dependency: str) -> list[str]:
        raise NotImplementedError("Method not overloaded.")



# src\file_conversor\utils\file.py

import os


class File:
    """
    File is a utility class for handling file operations.
    """

    def __init__(self, path: str) -> None:
        """
        Initialize the File object with a given path.

        :param path: The path to the file or directory.
        :raises ValueError: If the path is not specified or is not a string.
        """
        super().__init__()

        if not path or not isinstance(path, str):
            raise ValueError("Path must be specified (string)")
        self.path = path

    def is_file(self) -> bool:
        """
        Check if a file exists at the given path.

        :return: True if the file exists, False otherwise.
        """
        return os.path.isfile(self.path)

    def is_dir(self) -> bool:
        """
        Check if a dir exists at the given path.

        :return: True if the dir exists, False otherwise.
        """
        return os.path.isdir(self.path)

    def mkdir(self) -> None:
        """
        Create a directory if it does not exist.

        :raises FileExistsError: If the path is a file.
        :raises FileNotFoundError: If the directory could not be created, or if path is a file.
        """
        if os.path.isfile(self.path):
            raise FileExistsError(f"Path '{self.path}' is a file, not a directory.")

        os.makedirs(self.path, exist_ok=True)
        if not os.path.isdir(self.path):
            raise FileNotFoundError(f"Directory '{self.path}' could not be created.")

    def check_supported_format(self, supported_formats: list | dict):
        """
        Check if the file has a supported format.

        :param supported_formats: List or dict of supported file formats.

        :raises ValueError: If the file format is not supported.
        """
        extension = self.get_extension()
        if extension not in supported_formats:
            raise ValueError(f"Unsupported format: {extension}. Supported formats are: {', '.join(supported_formats)}")

    def get_extension(self) -> str:
        """
        Get the file extension.

        :return: The file extension.
        """
        return (os.path.splitext(self.path)[1])[1:].lower()

    def get_filename(self) -> str:
        """
        Get the file name without the extension.

        :return: The file name without the extension.
        """
        return os.path.splitext(os.path.basename(self.path))[0]

    def get_dirname(self) -> str:
        """
        Get the directory name of the file.

        :return: The directory name.
        """
        return os.path.dirname(self.path)

    def get_full_dirname(self) -> str:
        """
        Get the full absolute path of the directory.

        :return: The full absolute path of the directory.
        """
        return os.path.abspath(self.get_dirname())

    def get_full_path(self) -> str:
        """
        Get the full absolute path of the file.

        :return: The full absolute path of the file.
        """
        return os.path.abspath(self.path)




# src/file_conversor/utils/__init__.py

"""
This module initializes the utils package.
It can contain utility functions or classes that are used across the application.
"""

from file_conversor.utils.file import File



# src\file_conversor\config\config.py

import json
import locale

from pathlib import Path
from typing import Any

from file_conversor.config.state import State


class Configuration:

    __instance = None

    @staticmethod
    def get_instance():
        if not Configuration.__instance:
            Configuration.__instance = Configuration()
        return Configuration.__instance

    def __init__(self) -> None:
        super().__init__()

        self.__config_path = State.get_resources_folder() / ".config.json"
        # Define configuration dictionary
        language = "en_US"
        if locale.getlocale() and locale.getlocale()[0]:
            language = locale.getlocale()[0]
        self.__data = {
            "language": language,    # Default: system language or "en_US"
            "install-deps": True,    # Default: ask user to confirm dependency installation
            "audio-bitrate": 192,    # Default audio bitrate in kbps
            "video-bitrate": 10000,  # Default video bitrate in kbps
            "image-quality": 90,     # Default image quality 90%
            "image-dpi": 200,        # Default image => PDF dpi
            "image-fit": 'into',     # Default image => PDF fit mode
            "image-page-size": None,  # Default image => PDF page size
            "image-set-metadata": True,  # Default image => PDF set metadata
            "pdf-compression": "medium",  # Default PDF compression level
            "pdf-preset": "1.5",         # Default PDF compatibility preset
        }

        self.load()

    def __repr__(self) -> str:
        return repr(self.__data)

    def __str__(self) -> str:
        return str(self.__data)

    def __getitem__(self, key) -> Any:
        return self.__data[key]

    def __setitem__(self, key, value):
        self.__data[key] = value

    def __delitem__(self, key):
        del self.__data[key]

    def __contains__(self, key):
        return key in self.__data

    def __len__(self):
        return len(self.__data)

    def to_dict(self):
        return self.__data.copy()

    def clear(self):
        self.__data.clear()
        self.load()

    def update(self, new: dict):
        self.__data.update(new)

    def load(self):
        """Load app configuration file"""
        if self.__config_path.exists():
            self.__data.update(json.loads(self.__config_path.read_text()))

    def save(self):
        """Save app configuration file"""
        self.__config_path.write_text(json.dumps(self.__data, indent=2))




# src\file_conversor\cli\svg_cmd.py

import typer

from pathlib import Path
from typing import Annotated, List

from rich import print
from rich.text import Text
from rich.panel import Panel
from rich.console import Group

# user-provided modules
from file_conversor.backend import PyMuSVGBackend

from file_conversor.config import Configuration, State, Log
from file_conversor.config.locale import get_translation

from file_conversor.utils import File
from file_conversor.utils.rich import get_progress_bar
from file_conversor.utils.validators import check_file_format, check_is_bool_or_none, check_valid_options

from file_conversor.system.win.ctx_menu import WinContextCommand, WinContextMenu

# get app config
CONFIG = Configuration.get_instance()
STATE = State.get_instance()
LOG = Log.get_instance()

_ = get_translation()
logger = LOG.getLogger(__name__)

svg_cmd = typer.Typer()


def register_ctx_menu(ctx_menu: WinContextMenu):
    icons_folder_path = State.get_icons_folder()
    # PyMuSVGBackend commands
    for ext in PyMuSVGBackend.SUPPORTED_IN_FORMATS:
        ctx_menu.add_extension(f".{ext}", [
            WinContextCommand(
                name="to_jpg",
                description="To JPG",
                command=f'{State.get_executable()} svg convert "%1" -o "%1.jpg"',
                icon=str(icons_folder_path / 'jpg.ico'),
            ),
            WinContextCommand(
                name="to_png",
                description="To PNG",
                command=f'{State.get_executable()} svg convert "%1" -o "%1.png"',
                icon=str(icons_folder_path / 'png.ico'),
            ),
        ])


# register commands in windows context menu
ctx_menu = WinContextMenu.get_instance()
ctx_menu.register_callback(register_ctx_menu)


# image convert
@svg_cmd.command(
    help=f"""
        {_('Convert a SVG file to a different format.')}
    """,
    epilog=f"""
        **{_('Examples')}:** 

        - `file_conversor svg convert input_file.svg -o output_file.jpg --dpi 300`

        - `file_conversor svg convert input_file.svg -o output_file.png`
    """)
def convert(
    input_file: Annotated[str, typer.Argument(
        help=f"{_('Input file')} ({', '.join(PyMuSVGBackend.SUPPORTED_IN_FORMATS)})",
        callback=lambda x: check_file_format(x, PyMuSVGBackend.SUPPORTED_IN_FORMATS, exists=True),
    )],

    output_file: Annotated[str, typer.Option("--output", "-o",
                                             help=f"{_('Output file')} ({', '.join(PyMuSVGBackend.SUPPORTED_OUT_FORMATS)})",
                                             callback=lambda x: check_file_format(x, PyMuSVGBackend.SUPPORTED_OUT_FORMATS),
                                             )],

    dpi: Annotated[int, typer.Option("--dpi", "-d",
                                     help=_("Image quality in dots per inch (DPI). Valid values are between 40-3600."),
                                     min=40, max=3600,
                                     )] = CONFIG["image-dpi"],
):
    pymusvg_backend = PyMuSVGBackend(verbose=STATE['verbose'])
    # display current progress
    with get_progress_bar() as progress:
        task = progress.add_task(f"{_('Processing file')} '{input_file}':", total=None)
        pymusvg_backend.convert(
            input_file=input_file,
            output_file=output_file,
            dpi=dpi,
        )
        progress.update(task, total=100, completed=100)
    logger.info(f"{_('File convertion')}: [green][bold]{_('SUCCESS')}[/bold][/green]")



# src\file_conversor\backend\abstract_backend.py

"""
This module provides functionalities for handling external backends.
"""

import os
import platform
import shutil
import subprocess
import typer

from pathlib import Path
from rich import print

# user-provided imports
from file_conversor.dependency import AbstractPackageManager

from file_conversor.config import Log
from file_conversor.config.locale import get_translation

LOG = Log.get_instance()

_ = get_translation()
logger = LOG.getLogger(__name__)


class AbstractBackend:
    """
    Class that provides an interface for handling internal/external backends.
    """

    @staticmethod
    def find_in_path(name: str | Path) -> Path:
        """
        Finds name path in PATH env

        :return: Path for name

        :raises FileNotFoundError: if name not found
        """
        path_str = shutil.which(name)
        if not path_str:
            raise FileNotFoundError(f"'{name}' {_('not found in PATH environment')}")
        path = Path(path_str).resolve()
        logger.info(f"'{name}' {_('found')}: {path}")
        return path

    @staticmethod
    def check_file_exists(filename: str):
        """
        Check if `filename` exists

        :raises FileNotFoundError: if file not found
        """
        if not os.path.isfile(filename):
            raise FileNotFoundError(f"{_("File")} '{filename}' {_("not found")}")

    @staticmethod
    def dump_streams(process: subprocess.Popen | None, stdout=True, stderr=True) -> str:
        """Dumps stdout and/or stderr into a string"""
        res = ""
        if not process:
            return res
        if process.stderr:
            for line in process.stderr:
                res += line
        if process.stdout:
            for line in process.stdout:
                res += line
        return res

    def __init__(
        self,
        pkg_managers: set[AbstractPackageManager] | None = None,
        install_answer: bool | None = None,
    ):
        """
        Initialize the abstract backend.

        Checks if external dependencies are installed, and if not, install them.

        :param pkg_managers: Pkg managers configured to install external dependencies. Defaults to None (no external dependency required).
        :param install_answer: If True, do not ask user to install dependency (auto install). If False, do not install missing dependencies. If None, ask user for action. Defaults to None.

        :raises RuntimeError: Cannot install missing dependency or unknown OS detected.
        """
        super().__init__()
        pkg_managers = pkg_managers if pkg_managers else set()

        # identify OS and package manager
        os_type = platform.system()
        for pkg_mgr in pkg_managers:
            if os_type not in pkg_mgr.get_supported_oses():
                continue
            # supported pkg manager found, proceed to check for dependencies
            missing_deps = pkg_mgr.check_dependencies()
            if not missing_deps:
                # no dependencies missing, skip
                break
            logger.warning(f"[bold]{_("Missing dependencies detected")}[/]: {", ".join(missing_deps)}")

            # install package manager, if not present already
            pkg_mgr_bin = pkg_mgr.get_pkg_manager_installed()
            if pkg_mgr_bin:
                logger.info(f"Package manager found in '{pkg_mgr_bin}'")
            else:
                user_prompt: bool
                if install_answer is None:
                    user_prompt = typer.confirm(
                        _("Install package manager for the current user?"),
                        default=True,
                    )
                else:
                    user_prompt = install_answer
                if user_prompt:
                    result = pkg_mgr.install_pkg_manager()
                    if result:
                        logger.info(f"Package manager installed in '{result}'")
                    logger.info(f"[bold]{_("Package Manager Installation")}[/]: [green]{_("SUCCESS")}[/]")

            # install missing dependencies
            if install_answer is None:
                user_prompt = typer.confirm(
                    _(f"Install missing dependencies for the current user?"),
                    default=True,
                )
            else:
                user_prompt = install_answer
            if user_prompt:
                pkg_mgr.install_dependencies(missing_deps)
                logger.info(f"[bold]{_("External Dependencies Installation")}[/]: [green]{_("SUCCESS")}[/]")



# src\file_conversor\system\lin\utils.py

import os
import platform

# Import only on Linux to avoid ImportError on other OSes
if platform.system() == "Linux":
    pass  # dummy, do nothing
else:
    pass  # Placeholder so the name exists


def is_admin() -> bool:
    """True if app running with admin priviledges, False otherwise."""
    return os.geteuid() == 0  # type: ignore


def reload_user_path():
    """Reload user PATH in current process."""
    # dummy, not needed in Linux
    pass




# src\file_conversor\cli\__init__.py

"""
This module initializes the CLI commands package.
"""

# office
from file_conversor.cli.office import doc_cmd, xls_cmd, ppt_cmd

# other commands
from file_conversor.cli.audio_video_cmd import audio_video_cmd
from file_conversor.cli.batch_cmd import batch_cmd
from file_conversor.cli.config_cmd import config_cmd
from file_conversor.cli.image_cmd import image_cmd
from file_conversor.cli.pdf_cmd import pdf_cmd
from file_conversor.cli.svg_cmd import svg_cmd
from file_conversor.cli.win_cmd import win_cmd



# src\file_conversor\config\state.py

import shutil
import sys

from pathlib import Path
from importlib.resources import files

from typing import Any

# user provided imports
from file_conversor.config.log import Log

# Get app config
LOG = Log.get_instance()

logger = LOG.getLogger(__name__)


# STATE ACTIONS
def disable_log(value):
    if not value:
        return
    logger.info(f"'File logging': [blue red]'DISABLED'[/]")
    LOG.set_dest_folder(None)


def disable_progress(value):
    if not value:
        return
    logger.info(f"Progress bars: [blue red]DISABLED[/]")


def enable_quiet_mode(value):
    if not value:
        return
    logger.info(f"Quiet mode: [blue bold]ENABLED[/]")
    LOG.set_level(Log.LEVEL_ERROR)


def enable_verbose_mode(value):
    if not value:
        return
    logger.info(f"Verbose mode: [blue bold]ENABLED[/]")
    LOG.set_level(Log.LEVEL_INFO)


def enable_debug_mode(value):
    if not value:
        return
    logger.info(f"Debug mode: [blue bold]ENABLED[/]")
    LOG.set_level(Log.LEVEL_DEBUG)


# STATE controller dict class
class State:
    __instance = None

    @staticmethod
    def get_executable() -> str:
        """Get the executable path for this app's CLI."""
        res = ""

        exe = shutil.which(sys.argv[0]) if sys.argv else None
        if exe and not exe.endswith(".py"):
            res = rf'"{exe}"'
        else:
            python_exe = sys.executable
            main_py = Path(rf"{State.get_resources_folder()}/__main__.py")
            res = rf'"{python_exe}" "{main_py}"'

        logger.debug(f"Executable cmd: {res}")
        return res

    @staticmethod
    def get_resources_folder() -> Path:
        """Get the absolute path of the included folders in pip."""
        res_path = Path(str(files("file_conversor"))).resolve()
        return res_path

    @staticmethod
    def get_icons_folder() -> Path:
        """Get the absolute path of the included folders in pip."""
        icons_path = State.get_resources_folder() / ".icons"
        logger.debug(f"Icons path: {icons_path}")
        return icons_path

    @staticmethod
    def get_locales_folder() -> Path:
        locales_path = State.get_resources_folder() / ".locales"
        logger.debug(f"Locales path: {locales_path}")
        return locales_path

    @staticmethod
    def get_instance():
        if not State.__instance:
            State.__instance = State()
        return State.__instance

    def __init__(self) -> None:
        super().__init__()
        self.__init_state()

    def __init_state(self):
        # Define state dictionary
        self.__data = {
            # app options
            "no-log": False,
            "no-progress": False,
            "quiet": False,
            "verbose": False,
            "debug": False,
        }
        self.__callbacks = {
            "no-log": disable_log,
            "no-progress": disable_progress,
            "quiet": enable_quiet_mode,
            "verbose": enable_verbose_mode,
            "debug": enable_debug_mode,
        }
        # run callbacks
        for key, value in self.__data.items():
            self._run_callbacks(key=key, value=value)

    def _run_callbacks(self, key: str, value):
        if key in self.__callbacks:
            self.__callbacks[key](value)

    def __repr__(self) -> str:
        return repr(self.__data)

    def __str__(self) -> str:
        return str(self.__data)

    def __getitem__(self, key) -> Any:
        if key not in self.__data:
            raise KeyError(f"Key '{key}' not found in STATE")
        return self.__data[key]

    def __setitem__(self, key, value):
        if key not in self.__data:
            raise KeyError(f"Key '{key}' is not a valid key for STATE. Valid options are {', '.join(self.__data.keys())}")
        self.__data[key] = value

        # run callback
        self._run_callbacks(key=key, value=value)

    def __contains__(self, key) -> bool:
        return key in self.__data

    def __len__(self) -> int:
        return len(self.__data)

    def update(self, new: dict):
        for key, value in new.items():
            self[key] = value



# src\file_conversor\backend\office\writer_backend.py

# user-provided imports
from file_conversor.config import Log
from file_conversor.config.locale import get_translation
from file_conversor.backend.office.abstract_libreoffice_backend import AbstractLibreofficeBackend

LOG = Log.get_instance()

_ = get_translation()
logger = LOG.getLogger(__name__)


class LibreofficeWriterBackend(AbstractLibreofficeBackend):
    """
    A class that provides an interface for handling doc files using ``writer`` (libreoffice).
    """

    SUPPORTED_IN_FORMATS = {
        "doc": {},
        "docx": {},
        "odt": {},
    }
    SUPPORTED_OUT_FORMATS = {
        "doc": {},
        "docx": {},
        "odt": {},
        "pdf": {},
        "html": {},
    }

    def __init__(
        self,
        verbose: bool = False,
    ):
        """
        Initialize the backend

        :param verbose: Verbose logging. Defaults to False.      
        """
        super().__init__(verbose=verbose)



"""
Module for LibreOffice backend (calc, writer, etc)
"""

from file_conversor.system import CURR_PLATFORM, PLATFORM_WINDOWS

from file_conversor.backend.office.calc_backend import LibreofficeCalcBackend
from file_conversor.backend.office.excel_backend import ExcelBackend

from file_conversor.backend.office.impress_backend import LibreofficeImpressBackend
from file_conversor.backend.office.powerpoint_backend import PowerPointBackend

from file_conversor.backend.office.writer_backend import LibreofficeWriterBackend
from file_conversor.backend.office.word_backend import WordBackend

if CURR_PLATFORM == PLATFORM_WINDOWS:
    DOC_BACKEND = WordBackend
    XLS_BACKEND = ExcelBackend
    PPT_BACKEND = PowerPointBackend
else:
    DOC_BACKEND = LibreofficeWriterBackend
    XLS_BACKEND = LibreofficeCalcBackend
    PPT_BACKEND = LibreofficeImpressBackend




# src\file_conversor\dependency\scoop_pkg_manager.py

import os
import shutil
import subprocess

from typing import Callable, Any, Iterable

# user-provided imports
from file_conversor.system import PLATFORM_WINDOWS

from file_conversor.config.log import Log
from file_conversor.config.locale import get_translation

from file_conversor.dependency.abstract_pkg_manager import AbstractPackageManager

_ = get_translation()
LOG = Log.get_instance()

logger = LOG.getLogger(__name__)


class ScoopPackageManager(AbstractPackageManager):
    def __init__(self,
                 dependencies: dict[str, str],
                 buckets: list[str] | None = None,
                 ) -> None:
        """
        Inits Scoop pkg manager.

        :param dependencies: Format {executable: dependency}
        :param buckets: Buckets to add, to allow dependency install.
        """
        super().__init__(dependencies)
        self._buckets = buckets if buckets else []

        self.add_pre_install_callback(self._ensure_buckets)

    def _get_pkg_manager_installed(self) -> str | None:
        return shutil.which("scoop")

    def _get_supported_oses(self) -> set[str]:
        return {PLATFORM_WINDOWS}

    def _get_cmd_install_pkg_manager(self) -> list[str]:
        return [
            "powershell",
            "-NoProfile",
            "-ExecutionPolicy", "Bypass",
            "-Command",
            "iwr -useb get.scoop.sh | iex"
        ]

    def _post_install_pkg_manager(self) -> None:
        # update current PATH (current process path)
        scoop_shims = os.path.expandvars(r"%USERPROFILE%\scoop\shims")
        os.environ["PATH"] += os.pathsep + scoop_shims

    def _get_cmd_install_dep(self, dependency: str) -> list[str]:
        pkg_mgr_bin = self._get_pkg_manager_installed()
        pkg_mgr_bin = pkg_mgr_bin if pkg_mgr_bin else "SCOOP_NOT_FOUND"
        return [pkg_mgr_bin, "install", dependency, "-k"]

    def _ensure_buckets(self):
        scoop_bin = self._get_pkg_manager_installed() or "SCOOP_NOT_FOUND"
        process = subprocess.run([scoop_bin, "bucket", "list"],
                                 capture_output=True,
                                 text=True, check=True)
        for bucket in self._buckets:
            if bucket in process.stdout:
                logger.info(f"{_('Skipping bucket')} '{bucket}'. {_('Already added in scoop')}.")
                continue
            logger.info(f"{_('Adding bucket')} '{bucket}' {_('to scoop')}.")
            subprocess.run([scoop_bin, "bucket", "add", bucket],
                           capture_output=True,
                           text=True, check=True)
            logger.info(f"{_('Bucket')} '{bucket}' {_('added to scoop')}.")



# src\file_conversor\backend\qpdf_backend.py

"""
This module provides functionalities for handling PDF files using ``qpdf`` backend.
"""

import shutil
import subprocess

# user-provided imports
from file_conversor.config.log import Log

from file_conversor.dependency import BrewPackageManager, ScoopPackageManager
from file_conversor.backend.abstract_backend import AbstractBackend

LOG = Log.get_instance()

logger = LOG.getLogger(__name__)


class QPDFBackend(AbstractBackend):
    """
    A class that provides an interface for handling PDF files using ``qpdf``.
    """

    SUPPORTED_IN_FORMATS = {
        "pdf": {},
    }
    SUPPORTED_OUT_FORMATS = {
        "pdf": {},
    }

    def __init__(
        self,
        install_deps: bool | None,
        verbose: bool = False,
    ):
        """
        Initialize the ``qpdf`` backend

        :param install_deps: Install external dependencies. If True auto install using a package manager. If False, do not install external dependencies. If None, asks user for action.

        :param verbose: Verbose logging. Defaults to False.      
        """
        super().__init__(
            pkg_managers={
                ScoopPackageManager({
                    "qpdf": "qpdf"
                }),
                BrewPackageManager({
                    "qpdf": "qpdf"
                }),
            },
            install_answer=install_deps,
        )
        self._verbose = verbose

        self._qpdf_bin = self.find_in_path("qpdf")

    def repair(
        self,
        input_file: str,
        output_file: str,
        decrypt_password: str | None = None,
        compress: bool = True,
    ) -> subprocess.Popen:
        """
        Repair input PDF file.

        :param input_files: Input PDF file. 
        :param output_files: Output PDF file.
        :param decryption_password: Decryption password for input PDF file. Defaults to None (do not decrypt).
        :param compress: Compress PDF output file structures. Defaults to True (compress structures, losslessly).

        :raises FileNotFoundError: if input file not found.
        """
        self.check_file_exists(input_file)

        # build command
        command = []
        command.extend([str(self._qpdf_bin)])
        if compress:
            command.extend(["--object-streams=generate", "--compress-streams=y", "--stream-data=compress"])
        if decrypt_password:
            command.extend([f"--password={decrypt_password}", "--decrypt"])
        command.extend(["--linearize", input_file, output_file])

        # Execute command
        logger.info("Executing QPDF ...")
        logger.debug(" ".join(command))
        process = subprocess.Popen(
            command,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True
        )
        return process




# src\file_conversor\cli\win_cmd.py

import typer

from typing import Annotated

from rich import print


# user-provided modules
from file_conversor.backend import WinRegBackend

from file_conversor.config import Configuration, State, Log
from file_conversor.config.locale import get_translation

from file_conversor.system.win import restart_explorer
from file_conversor.system.win.ctx_menu import WinContextMenu

# get app config
CONFIG = Configuration.get_instance()
STATE = State.get_instance()
LOG = Log.get_instance()

_ = get_translation()
logger = LOG.getLogger(__name__)

win_cmd = typer.Typer()

# PANELS
CONTEXT_MENU_PANEL = _("Context menu")


# win install-menu
@win_cmd.command(
    rich_help_panel=CONTEXT_MENU_PANEL,
    help=f"""
        {_('Installs app context menu (right click in Windows Explorer).')}        
    """,
    epilog=f"""
**{_('Examples')}:** 

- `file_conversor win install-menu` 
""")
def install_menu():
    winreg_backend = WinRegBackend(verbose=STATE["verbose"])

    logger.info(f"{_('Installing app context menu in Windows Explorer')} ...")

    # Define registry path
    ctx_menu = WinContextMenu.get_instance()
    # logger.debug("---- .REG file contents ----")
    # logger.debug(repr(ctx_menu.get_reg_file()))

    winreg_backend.import_file(ctx_menu.get_reg_file())

    logger.info(f"{_('Restarting explorer.exe')} ...")
    restart_explorer()
    logger.info(f"{_('Context Menu Install')}: [bold green]{_('SUCCESS')}[/].")


# win uninstall-menu
@win_cmd.command(
    rich_help_panel=CONTEXT_MENU_PANEL,
    help=f"""
        {_('Uninstalls app context menu (right click in Windows Explorer).')}        
    """,
    epilog=f"""
**{_('Examples')}:** 

- `file_conversor win uninstall-menu` 
""")
def uninstall_menu():
    winreg_backend = WinRegBackend(verbose=STATE["verbose"])

    logger.info(f"{_('Removing app context menu from Windows Explorer')} ...")

    # Define registry path
    ctx_menu = WinContextMenu.get_instance()
    # logger.debug("---- .REG file contents ----")
    # logger.debug(repr(ctx_menu.get_reg_file()))

    winreg_backend.delete_keys(ctx_menu.get_reg_file())

    logger.info(f"{_('Context Menu Uninstall')}: [bold green]{_('SUCCESS')}[/].")



# src\file_conversor\system\win\__init__.py

# exports
from file_conversor.system.win.ctx_menu import WinContextMenu, WinContextCommand
from file_conversor.system.win.reg import WinRegFile, WinRegKey
from file_conversor.system.win.utils import is_admin, reload_user_path, restart_explorer



# src\file_conversor\backend\pypdf_backend.py

"""
This module provides functionalities for handling PDF files using ``pypdf`` backend.
"""

import math

from pypdf import PdfReader, PdfWriter
from pypdf.constants import UserAccessPermissions

from typing import Iterable

# user-provided imports
from file_conversor.config import Log

from file_conversor.backend.abstract_backend import AbstractBackend

LOG = Log.get_instance()

logger = LOG.getLogger(__name__)


class PyPDFBackend(AbstractBackend):
    """
    A class that provides an interface for handling PDF files using ``pypdf``.
    """

    SUPPORTED_IN_FORMATS = {
        "pdf": {},
    }
    SUPPORTED_OUT_FORMATS = {
        "pdf": {},
    }

    def __init__(
        self,
        verbose: bool = False,
    ):
        """
        Initialize the ``pypdf`` backend

        :param verbose: Verbose logging. Defaults to False.      
        """
        super().__init__()
        self._verbose = verbose

    def len(self,
            input_file: str,
            ) -> int:
        """
        Get number of pages of input file.

        :param input_file: Input PDF file.

        :return: Number of pages.

        :raises FileNotFoundError: if input file not found
        """
        self.check_file_exists(input_file)

        with PdfReader(input_file) as reader:
            return len(reader.pages)

    def merge(self,
              output_file: str,
              input_files: dict[str, str],
              ):
        """
        Merge input files into an output.

        ```python
        input_files = {
            file_path_1: password_1,
            file_path_2: password_2,
            ...
        }
        ```

        :param output_file: Output PDF file.
        :param input_files: Input PDF files. Format ``dict[file_path, password]`` (if file not encrypted, use ``password = None``).

        :raises FileNotFoundError: if input file not found.
        """
        output_file = output_file.replace(".pdf", "")

        with PdfWriter() as writer:
            for in_file, decrypt_password in input_files.items():
                self.check_file_exists(in_file)
                with PdfReader(in_file) as reader:
                    if decrypt_password and reader.is_encrypted:
                        reader.decrypt(decrypt_password)
                    for page in reader.pages:
                        writer.add_page(page)
            writer.write(f"{output_file}.pdf")

    def split(self,
              output_file: str,
              input_file: str,
              decrypt_password: str | None = None,
              ):
        """
        Split input files into 1-page PDF output files.

        :param output_file: Output PDF file (will be appended with ``_X.pdf`` where ``X`` is the page number)
        :param input_file: Input PDF file

        :param decrypt_password: Password used to decrypt file, if needed. Defaults to None (do not decrypt).

        :raises FileNotFoundError: if input file not found
        """
        self.check_file_exists(input_file)
        output_file = output_file.replace(".pdf", "")

        with PdfReader(input_file) as reader:
            if decrypt_password and reader.is_encrypted:
                reader.decrypt(decrypt_password)
            for i, page in enumerate(reader.pages):
                with PdfWriter() as writer:
                    writer.add_page(page)
                    writer.write(f"{output_file}_{i + 1}.pdf")

    def extract(self,
                output_file: str,
                input_file: str,
                pages: Iterable[int],
                decrypt_password: str | None = None,
                ):
        """
        Extract specific pages from input files into a PDF output file.

        :param output_file: Output PDF file
        :param input_file: Input PDF file

        :param pages: List of pages to extract (0-indexed).
        :param decrypt_password: Password used to decrypt file, if needed. Defaults to None (do not decrypt).

        :raises FileNotFoundError: if input file not found
        """
        self.check_file_exists(input_file)
        output_file = output_file.replace(".pdf", "")

        with PdfReader(input_file) as reader, PdfWriter() as writer:
            if decrypt_password and reader.is_encrypted:
                reader.decrypt(decrypt_password)

            for page_num in pages:
                writer.add_page(reader.pages[page_num])
            writer.write(f"{output_file}.pdf")

    def rotate(self,
               output_file: str,
               input_file: str,
               rotations: dict[int, int],
               decrypt_password: str | None = None,
               ):
        """
        Rotate specific pages from input files, generating a PDF output file.

        :param output_file: Output PDF file.
        :param input_file: Input PDF file.

        :param rotations: Dict format { page_num: rotation_degrees }. Degree must be 0 or multiples of 90 degrees - positive or negative.
        :param decrypt_password: Password used to decrypt file, if needed. Defaults to None (do not decrypt).

        :raises FileNotFoundError: if input file not found
        :raises ValueError: if rotation degree is invalid (valid values are 0 or multiples of 90 degrees - positive or negative).        
        """
        self.check_file_exists(input_file)
        output_file = output_file.replace(".pdf", "")

        with PdfReader(input_file) as reader, PdfWriter() as writer:
            if decrypt_password and reader.is_encrypted:
                reader.decrypt(decrypt_password)

            for i, page in enumerate(reader.pages):
                rotation = int(rotations.get(i, 0))

                # parse rotation argument
                rotation = int(math.fmod(rotation, 360))
                if rotation < 0:
                    rotation += 360  # fix rotation signal

                if rotation not in (0, 90, 180, 270):
                    raise ValueError(f"Page '{i}' rotation {rotation} is invalid. Rotation must be 0, 90, 180 or 270 degrees.")

                # execute page rotation
                logger.debug(f"Rotating page {i} by {rotation} deg")
                if rotation > 0:
                    page.rotate(rotation)  # clockwise: 90, 180, 270
                writer.add_page(page)
            # save output file
            writer.write(f"{output_file}.pdf")

    def encrypt(self,
                output_file: str,
                input_file: str,
                owner_password: str,
                user_password: str | None = None,
                decrypt_password: str | None = None,
                permission_annotate: bool = False,
                permission_fill_forms: bool = False,
                permission_modify: bool = False,
                permission_modify_pages: bool = False,
                permission_copy: bool = False,
                permission_accessibility: bool = True,
                permission_print_low_quality: bool = True,
                permission_print_high_quality: bool = True,
                permission_all: bool = False,
                encryption_algorithm: str = "AES-256",
                ):
        """
        Encrypt input file, generating a protected PDF output file.

        **Available permissions:**
        - ``permission_annotate``: User can add/modify annotations (comments, highlight text, etc) and interactive forms. Default to False (not set).
        - ``permission_fill_forms``: User can fill form fields (subset permission of `permission_annotate`). Default to False (not set).
        - ``permission_modify``: User can modify the document (e.g., add / edit text, add / edit images, etc). Default to False (not set).
        - ``permission_modify_pages``: User can insert, delete, or rotate pages (subset of ``permission_modify``). Default to False (not set).
        - ``permission_copy``: User can copy text/images. Default to False (not set).
        - ``permission_accessibility``: User can use screen readers for accessibility. Default to True (not set).
        - ``permission_print_low_quality``: User can PRINT pdf (low quality). Defaults to True (allow).
        - ``permission_print_high_quality``: User can PRINT pdf (high quality). Requires `permission_print_low_quality=True`. Defaults to True (allow).
        - ``permission_all``: User has ALL PERMISSIONS. If True, it overrides all other permissions. Defaults to False (not set).

        **Encryption algorithms:**
        - ``AES-256``    (high security, compatible with 2008+ PDF readers - **RECOMMENDED**)
        - ``AES-128``    (good security, good compatibility - **RECOMMENDED**)
        - ``AES-256-R5`` (high security,  lowest compatibility - not recommended)
        - ``RC4-128``    (weak security, highest compatibility - not recommended)
        - ``RC4-40``     (weakest security - not recommended)

        :param output_file: Output PDF file.
        :param input_file: Input PDF file.

        :param owner_password: Owner password for encryption. Owner has ALL PERMISSIONS in the output PDF file.
        :param user_password:  User password for encryption. User has ONLY THE PERMISSIONS specified in the arguments. Defaults to None (user and owner password are the same).
        :param decrypt_password: Password used to decrypt file, if needed. Defaults to None (do not decrypt).

        :param permission_annotate: User can add/modify annotations (comments, highlight text, etc) and interactive forms. Default to False (not set).
        :param permission_fill_forms: User can fill form fields (subset permission of `permission_annotate`). Default to False (not set).
        :param permission_modify: User can modify the document (e.g., add / edit text, add / edit images, etc). Default to False (not set).
        :param permission_modify_pages: User can insert, delete, or rotate pages (subset of ``permission_modify``). Default to False (not set).
        :param permission_copy: User can copy text/images. Default to False (not set).
        :param permission_accessibility: User can use screen readers for accessibility. Default to True (allow).
        :param permission_print_low_quality: User can PRINT pdf (low quality). Defaults to True (allow).
        :param permission_print_high_quality: User can PRINT pdf (high quality). Requires `permission_print_low_quality=True`. Defaults to True (allow).
        :param permission_all: User has ALL PERMISSIONS. If True, it overrides all other permissions. Defaults to False (not set).

        :param encryption_algorithm: Encryption algorithm used. Valid options are "RC4-40", "RC4-128", "AES-128", "AES-256-R5", or "AES-256". Defaults to "AES-256" (for enhanced security and compatibility).

        :raises FileNotFoundError: if input file not found
        """
        self.check_file_exists(input_file)
        output_file = output_file.replace(".pdf", "")

        with PdfReader(input_file) as reader, PdfWriter() as writer:
            if decrypt_password and reader.is_encrypted:
                reader.decrypt(decrypt_password)

            if encryption_algorithm not in (None, "RC4-40", "RC4-128", "AES-128", "AES-256-R5", "AES-256"):
                raise ValueError(f"Encryption algorithm '{encryption_algorithm}' is invalid.  Valid options are 'RC4-40', 'RC4-128', 'AES-128', 'AES-256-R5', or 'AES-256'.")

            # set user permissions
            permissions = UserAccessPermissions(0)
            if permission_annotate:
                permissions |= UserAccessPermissions.ADD_OR_MODIFY
            if permission_fill_forms:
                permissions |= UserAccessPermissions.FILL_FORM_FIELDS
            if permission_modify:
                permissions |= UserAccessPermissions.MODIFY
            if permission_modify_pages:
                permissions |= UserAccessPermissions.ASSEMBLE_DOC
            if permission_copy:
                permissions |= UserAccessPermissions.EXTRACT
            if permission_accessibility:
                permissions |= UserAccessPermissions.EXTRACT_TEXT_AND_GRAPHICS
            if permission_print_low_quality:
                permissions |= UserAccessPermissions.PRINT
            if permission_print_high_quality:
                permissions |= UserAccessPermissions.PRINT_TO_REPRESENTATION
            if permission_all:
                permissions |= UserAccessPermissions.all()

            # add pages into writer
            for page in reader.pages:
                writer.add_page(page)

            # Encrypt with user and owner passwords
            writer.encrypt(
                owner_password=owner_password,
                user_password=user_password if user_password else owner_password,
                permissions_flag=permissions,
                algorithm=encryption_algorithm,
                use_128bit=True,
            )

            # save output file
            writer.write(f"{output_file}.pdf")

    def decrypt(self,
                output_file: str,
                input_file: str,
                password: str,
                ):
        """
        Decrypt input file, generating a non-protected PDF output file.

        :param output_file: Output PDF file.
        :param input_file: Input PDF file.

        :param password: Password used to decrypt file.

        :raises FileNotFoundError: if input file not found
        """
        self.check_file_exists(input_file)
        output_file = output_file.replace(".pdf", "")

        with PdfReader(input_file) as reader, PdfWriter() as writer:
            # decrypt file
            if reader.is_encrypted:
                reader.decrypt(password)

            # Generate decrypted pdf
            for page in reader.pages:
                writer.add_page(page)
            writer.write(f"{output_file}.pdf")



# src\file_conversor\utils\formatters.py

def format_bytes(size: float) -> str:
    """Format size in bytes, KB, MB, GB, or TB"""
    # Tamanho em bytes para string legível
    for unit in ['bytes', 'KB', 'MB', 'GB', 'TB']:
        if size < 1024.0:
            return f"{size:.1f} {unit}"
        size /= 1024.0
    return f"{size:.1f} TB"


def format_bitrate(bps: int) -> str:
    """Format bitrate in bps, kbps or Mbps"""
    if bps >= 1_000_000:
        return f"{bps / 1_000_000:.2f} Mbps"
    elif bps >= 1000:
        return f"{bps / 1000:.0f} kbps"
    return f"{bps} bps"



# src\file_conversor\backend\office\word_backend.py

from pathlib import Path
from typing import Iterable

# user-provided imports
from file_conversor.system import CURR_PLATFORM, PLATFORM_WINDOWS
from file_conversor.config import Log
from file_conversor.config.locale import get_translation
from file_conversor.backend.abstract_backend import AbstractBackend

# conditional import
if CURR_PLATFORM == PLATFORM_WINDOWS:
    from win32com import client
else:
    client = None

LOG = Log.get_instance()

_ = get_translation()
logger = LOG.getLogger(__name__)


class WordBackend(AbstractBackend):
    """
    A class that provides an interface for handling doc files using ``word`` (comtypes).
    """

    SUPPORTED_IN_FORMATS = {
        "doc": {},
        "docx": {},
        "odt": {},
    }
    SUPPORTED_OUT_FORMATS = {
        # format = wdFormat VBA code
        # https://learn.microsoft.com/en-us/office/vba/api/word.wdsaveformat
        "doc": {'format': 0},
        "docx": {'format': 16},
        "odt": {'format': 23},
        "pdf": {'format': 17},
        "html": {'format': 8},
    }

    def __init__(
        self,
        verbose: bool = False,
    ):
        """
        Initialize the backend

        :param verbose: Verbose logging. Defaults to False.      
        """
        super().__init__()
        self._verbose = verbose

    def convert(
        self,
        output_file: str,
        input_file: str,
    ):
        """
        Convert input file into an output file.

        :param output_file: Output file.
        :param input_file: Input file.        

        :raises FileNotFoundError: if input file not found.
        :raises RuntimeError: if os != Windows.
        """
        if not client:
            raise RuntimeError("This backend supports only Windows OS")
        input_path = Path(input_file).resolve()
        output_path = Path(output_file).resolve()

        self.check_file_exists(str(input_path))

        out_config = WordBackend.SUPPORTED_OUT_FORMATS[output_path.suffix[1:]]

        word = client.Dispatch("Word.Application")
        doc = word.Documents.Open(str(input_path))
        doc.SaveAs(str(output_path),
                   FileFormat=out_config['format'],
                   )
        doc.Close()
        try:
            word.Quit()
        except:
            logger.warning(f"{_('Failed to close Word properly')}.")



# src\file_conversor\system\win\reg.py

import json
import re

from pathlib import Path
from typing import Iterable, Self

# user-provided imports
from file_conversor.config.log import Log

LOG = Log.get_instance()

logger = LOG.getLogger(__name__)


class WinRegKey:
    KEY_RE = re.compile(r'^\[(.*?)\]$')
    VALUE_RE = re.compile(r'^(?:"([^"]+)"|@)=(?:"(.*)"|dword:[0-9a-fA-F]+|hex:[0-9a-fA-F,]+)$')

    def __init__(self, path: str) -> None:
        r"""Initializes RegKey with path of type HKEY_*\\* """
        super().__init__()
        self._data: dict[str, str] = {}
        self.path = path

    @property
    def path(self):
        return self._path

    @path.setter
    def path(self, p):
        self._path = p

    def __eq__(self, other):
        return isinstance(other, WinRegKey) and self.path == other.path

    def __hash__(self):
        return hash(self.path)

    def __str__(self) -> str:
        return self.path

    def __repr__(self) -> str:
        res = [f"[{self._path}]"]
        for name, content in self._data.items():
            name_str = json.dumps(name) if name and name != "@" else "@"
            content_str = json.dumps(content) if content else '""'
            res.append(f'{name_str}={content_str}')
        return "\n".join(res)

    def __contains__(self, name: str):
        return name in self._data

    def __getitem__(self, name: str):
        return self._data[name]

    def items(self):
        return self._data.items()

    def add_value(self, name: str, content: str | int | bytes) -> Self:
        """
        Add value to windows key

        :param name: value name
        :param content: value content.

        :raises TypeError: content type is invalid
        """
        name_str = "@" if not name else name

        content_str = ""
        if isinstance(content, str):
            content_str = content
        elif isinstance(content, int):
            content_str = f"dword:{content:08x}"  # Pad to 8 digits
        elif isinstance(content, bytes):
            content_str = f"hex:{content.hex()}"
        else:
            raise TypeError(f"Content '{content}' type is not str|int.")

        self._data[name_str] = content_str
        # logger.debug(f'Add value "{name_str}={content_str}"')
        return self

    def del_value(self, name: str) -> Self:
        if name in self._data:
            del self._data[name]
            # logger.debug(f'Del value "{name}"')
        return self

    def update(self, key: Self | dict[str, str | int | bytes]) -> Self:
        if isinstance(key, WinRegKey):
            self._data.update(key._data)
        else:
            for name, content in key.items():
                self.add_value(name, content)
        # logger.info(f'Updated reg key {self.path}')
        return self


class WinRegFile:
    def __init__(self, input_path_or_winregfile: str | Path | Self | None = None) -> None:
        """
        Initialized WinRegFile

        :param input_path_or_winregfile: .REG file or WinRegFile. Defaults to None (do not load file).
        """
        super().__init__()
        self._data: dict[str, WinRegKey] = {}

        if isinstance(input_path_or_winregfile, (str, Path)):
            self.load(input_path_or_winregfile)
        elif isinstance(input_path_or_winregfile, WinRegFile):
            self.update(input_path_or_winregfile)

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        res = ["Windows Registry Editor Version 5.00"]
        for name, key in self._data.items():
            res.append(repr(key))
        return "\n\n".join(res)

    def __iter__(self):
        return self._data.__iter__()

    def __getitem__(self, key: str | WinRegKey):
        if isinstance(key, str):
            return self._data[key]
        elif isinstance(key, WinRegKey):
            return self._data[key.path]
        else:
            raise KeyError("Key search supports only str|WinRegKey")

    def items(self):
        return self._data.items()

    def add_key(self, key: WinRegKey) -> Self:
        if key.path in self._data:
            key_current = self._data[key.path]
            for name, value in key.items():
                key_current.add_value(name, value)
        else:
            self._data[key.path] = key
        # logger.debug(f"Add key '{key.path}'")
        return self

    def del_key(self, key: WinRegKey) -> Self:
        if key.path in self._data:
            del self._data[key.path]
            # logger.debug(f"Delete key '{key.path}'")
        return self

    def update(self, reg: Self | Iterable[WinRegKey]) -> Self:
        if isinstance(reg, WinRegFile):
            for _, key in reg.items():
                self.add_key(key)
        else:
            for key in reg:
                self.add_key(key)
        # logger.info(f"Updated reg file")
        return self

    def dump(self, output_file: str | Path):
        """
        Dump WinRegFile into .REG file.

        :param output_file: Output .REG file.     
        """
        out_file = f"{str(output_file).replace(".reg", "")}.reg"
        with open(out_file, "wb") as f:
            f.write(b'\xff\xfe')     # BOM
            f.write(repr(self).encode("utf-16le"))
        logger.info(f"Dumped reg file into '{output_file}'")

    def dumps(self) -> str:
        """
        Dumps WinRegFile into string.
        """
        return repr(self)

    def load(self, input_path: str | Path):
        """
        Loads a .REG file into WinRegFile

        :param input_path: Input .REG file

        :raises FileNotFoundError: Input file not exists.
        :raises RuntimeError: Invalid REG format. Value not associated with key.
        """
        input_path = Path(input_path)
        if not input_path.exists():
            raise FileNotFoundError(f"Input file '{input_path}' not exists")
        input_text = input_path.read_text(encoding="utf-16")
        if not input_text.startswith("Windows Registry Editor Version 5.00"):
            raise RuntimeError("Invalid .REG format")

        reg_key: WinRegKey | None = None
        for line in input_text.splitlines():
            match_key = re.match(WinRegKey.KEY_RE, line)
            if match_key:
                reg_key = WinRegKey(match_key.group(1))
                self.add_key(reg_key)
                continue

            match_value = re.match(WinRegKey.VALUE_RE, line)
            if match_value:
                name = match_value.group(1) or "@"
                content = match_value.group(2)

                if content.startswith("dword:"):
                    content = int(content.replace("dword:", ""), 16)
                elif content.startswith("hex:"):
                    hex_str = content.replace("hex:", "").replace(",", "")
                    content = bytes.fromhex(hex_str)
                if not reg_key:
                    raise RuntimeError(f"Invalid reg file. Value '{name}' not associated with a key.")
                reg_key.add_value(name, content)

        logger.info(f"Input file '{input_path}' loaded")




# src\file_conversor\cli\pdf_cmd.py

import re
import time
import typer

from pathlib import Path
from typing import Annotated, List

from rich import print

# user-provided modules
from file_conversor.backend import PyPDFBackend, QPDFBackend, PyMuPDFBackend, GhostscriptBackend

from file_conversor.config import Configuration, State, Log
from file_conversor.config.locale import get_translation

from file_conversor.utils.rich import get_progress_bar
from file_conversor.utils.validators import check_file_format, check_valid_options

from file_conversor.system.win.ctx_menu import WinContextCommand, WinContextMenu

# get app config
CONFIG = Configuration.get_instance()
STATE = State.get_instance()
LOG = Log.get_instance()

_ = get_translation()
logger = LOG.getLogger(__name__)

# typer PANELS
SECURITY_PANEL = _(f"Security commands")

pdf_cmd = typer.Typer()


def register_ctx_menu(ctx_menu: WinContextMenu):
    icons_folder_path = State.get_icons_folder()
    ctx_menu.add_extension(".pdf", [
        WinContextCommand(
            name="to_png",
            description="To PNG",
            command=f'{State.get_executable()} pdf convert "%1" -o "%1.png"',
            icon=str(icons_folder_path / 'png.ico'),
        ),
        WinContextCommand(
            name="to_jpg",
            description="To JPG",
            command=f'{State.get_executable()} pdf convert "%1" -o "%1.jpg"',
            icon=str(icons_folder_path / 'jpg.ico'),
        ),
        WinContextCommand(
            name="compress",
            description="Compress",
            command=f'{State.get_executable()} pdf compress "%1"',
            icon=str(icons_folder_path / 'compress.ico'),
        ),
        WinContextCommand(
            name="repair",
            description="Repair",
            command=f'{State.get_executable()} pdf repair "%1"',
            icon=str(icons_folder_path / 'repair.ico'),
        ),
        WinContextCommand(
            name="split",
            description="Split",
            command=f'{State.get_executable()} pdf split "%1"',
            icon=str(icons_folder_path / 'split.ico'),
        ),
        WinContextCommand(
            name="extract",
            description="Extract",
            command=f'cmd /k "{State.get_executable()} pdf extract "%1""',
            icon=str(icons_folder_path / 'extract.ico'),
        ),
        WinContextCommand(
            name="rotate_anticlock_90",
            description="Rotate Left",
            command=f'{State.get_executable()} pdf rotate "%1" -r "1-:-90"',
            icon=str(icons_folder_path / "rotate_left.ico"),
        ),
        WinContextCommand(
            name="rotate_clock_90",
            description="Rotate Right",
            command=f'{State.get_executable()} pdf rotate "%1" -r "1-:90"',
            icon=str(icons_folder_path / "rotate_right.ico"),
        ),
        WinContextCommand(
            name="encrypt",
            description="Encrypt",
            command=f'cmd /k "{State.get_executable()} pdf encrypt "%1""',
            icon=str(icons_folder_path / "padlock_locked.ico"),
        ),
        WinContextCommand(
            name="decrypt",
            description="Decrypt",
            command=f'cmd /k "{State.get_executable()} pdf decrypt "%1""',
            icon=str(icons_folder_path / "padlock_unlocked.ico"),
        ),
    ])


# register commands in windows context menu
ctx_menu = WinContextMenu.get_instance()
ctx_menu.register_callback(register_ctx_menu)


# pdf repair
@pdf_cmd.command(
    help=f"""
        {_('Repair (lightly) corrupted PDF files (requires QPDF external library).')}        
    """,
    epilog=f"""
**{_('Examples')}:** 

- `file_conversor pdf repair input_file.pdf -o output_file.pdf` 
""")
def repair(
    input_file: Annotated[str, typer.Argument(help=f"{_('Input file')} ({', '.join(QPDFBackend.SUPPORTED_IN_FORMATS)})",
                                              callback=lambda x: check_file_format(x, QPDFBackend.SUPPORTED_IN_FORMATS, exists=True),
                                              )],
    output_file: Annotated[str | None, typer.Option("--output", "-o",
                                                    help=f"{_('Output file')} ({', '.join(QPDFBackend.SUPPORTED_OUT_FORMATS)}). {_('Defaults to None')} ({_('use the same input file as output name')}, {_('with _repaired.pdf at the end)')}.",
                                                    callback=lambda x: check_file_format(x, QPDFBackend.SUPPORTED_OUT_FORMATS),
                                                    )] = None,
    decrypt_password: Annotated[str | None, typer.Option("--password", "-p",
                                                         help=_("Password used to open protected file. Defaults to None (do not decrypt)."),
                                                         )] = None,
    compress: Annotated[bool, typer.Option("--compress", "-c",
                                           help=_("Compress output PDF file (losslessly). Defaults to True (compress)."),
                                           is_flag=True,
                                           )] = True,
):
    qpdf_backend = QPDFBackend(verbose=STATE["verbose"], install_deps=CONFIG['install-deps'])
    with get_progress_bar() as progress:
        task1 = progress.add_task(f"{_('Processing file')} '{input_file}':", total=None,)

        process = qpdf_backend.repair(
            # files
            input_file=input_file,
            output_file=output_file if output_file else f"{input_file.replace(".pdf", "")}_repaired.pdf",

            # options
            decrypt_password=decrypt_password,
            compress=compress,
        )
        while process.poll() is None:
            time.sleep(0.25)
        progress.update(task1, total=100, completed=100)

    process.wait()
    if process.returncode != 0:
        raise RuntimeError(qpdf_backend.dump_streams(process))

    logger.info(f"{_('Repair PDF')}: [bold green]{_('SUCCESS')}[/].")


# pdf compress
@pdf_cmd.command(
    help=f"""
        {_('Compress a PDF file (requires Ghostscript external library).')}
    """,
    epilog=f"""
        **{_('Examples')}:** 

        - `file_conversor pdf compress input_file.pdf -o output_file.pdf`

        - `file_conversor pdf compress input_file.pdf -o output_file.pdf -c high`

        - `file_conversor pdf compress input_file.pdf`
    """)
def compress(
    input_file: Annotated[str, typer.Argument(
        help=f"{_('Input file')} ({', '.join(GhostscriptBackend.SUPPORTED_IN_FORMATS)})",
        callback=lambda x: check_file_format(x, GhostscriptBackend.SUPPORTED_IN_FORMATS, exists=True),
    )],

    output_file: Annotated[str | None, typer.Option("--output", "-o",
                                                    help=f"{_('Output file')} ({', '.join(GhostscriptBackend.SUPPORTED_OUT_FORMATS)}). {_('Defaults to None')} ({_('use the same input file as output name')}, {_('with _compressed.pdf at the end)')}",
                                                    callback=lambda x: check_file_format(x, GhostscriptBackend.SUPPORTED_OUT_FORMATS),
                                                    )] = None,

    compression: Annotated[str, typer.Option("--compression", "-c",
                                             help=f"{_('Compression level (high compression = low quality). Valid values are')} {', '.join(["low", "medium", "high", "none"])}. {_('Defaults to')} {CONFIG["pdf-compression"]}.",
                                             callback=lambda x: check_valid_options(x, ["low", "medium", "high", "none"]),
                                             )] = CONFIG["pdf-compression"],

    preset: Annotated[str, typer.Option("--preset", "-p",
                                        help=f"{_('Compatibility preset. Valid values are')} '1.3', '1.4', ..., '1.7' . {_('Defaults to')} {CONFIG["pdf-preset"]}.",
                                        callback=lambda x: check_valid_options(x, ["1.3", "1.4", "1.5", "1.6", "1.7"]),
                                        )] = CONFIG["pdf-preset"],
):
    gs_backend = GhostscriptBackend(
        install_deps=CONFIG['install-deps'],
        verbose=STATE['verbose'],
    )

    compression_level = "NOT_SET"
    if compression == "none":
        compression_level = GhostscriptBackend.COMPRESSION_NONE
    elif compression == "low":
        compression_level = GhostscriptBackend.COMPRESSION_LOW
    elif compression == "medium":
        compression_level = GhostscriptBackend.COMPRESSION_MEDIUM
    elif compression == "high":
        compression_level = GhostscriptBackend.COMPRESSION_HIGH

    # display current progress
    with get_progress_bar() as progress:
        task = progress.add_task(f"{_('Processing file')} '{input_file}':", total=None)
        gs_backend.compress(
            input_file=input_file,
            output_file=output_file if output_file else f"{input_file.replace(".pdf", "")}_compressed.pdf",
            compression_level=compression_level,
            compatibility_preset=preset,
        )
        progress.update(task, total=100, completed=100)
    logger.info(f"{_('File convertion')}: [green][bold]{_('SUCCESS')}[/bold][/green]")


# pdf convert
@pdf_cmd.command(
    help=f"""
        {_('Convert a PDF file to a different format.')}
    """,
    epilog=f"""
        **{_('Examples')}:** 

        - `file_conversor pdf convert input_file.pdf -o output_file.jpg --dpi 200`

        - `file_conversor pdf convert input_file.pdf -o output_file.png`
    """)
def convert(
    input_file: Annotated[str, typer.Argument(
        help=f"{_('Input file')} ({', '.join(PyMuPDFBackend.SUPPORTED_IN_FORMATS)})",
        callback=lambda x: check_file_format(x, PyMuPDFBackend.SUPPORTED_IN_FORMATS, exists=True),
    )],

    output_file: Annotated[str, typer.Option("--output", "-o",
                                             help=f"{_('Output file')} ({', '.join(PyMuPDFBackend.SUPPORTED_OUT_FORMATS)})",
                                             callback=lambda x: check_file_format(x, PyMuPDFBackend.SUPPORTED_OUT_FORMATS),
                                             )],

    dpi: Annotated[int, typer.Option("--dpi", "-d",
                                     help=_("Image quality in dots per inch (DPI). Valid values are between 40-3600."),
                                     min=40, max=3600,
                                     )] = CONFIG["image-dpi"],
):
    pymupdf_backend = PyMuPDFBackend(verbose=STATE['verbose'])
    # display current progress
    with get_progress_bar() as progress:
        task = progress.add_task(f"{_('Processing file')} '{input_file}':", total=None)
        pymupdf_backend.convert(
            input_file=input_file,
            output_file=output_file,
            dpi=dpi,
        )
        progress.update(task, total=100, completed=100)
    logger.info(f"{_('File convertion')}: [green][bold]{_('SUCCESS')}[/bold][/green]")


# pdf merge
@pdf_cmd.command(
    help=f"""
        {_('Merge (join) input PDFs into a single PDF file.')}
    """,
    epilog=f"""
**{_('Examples')}:** 



*{_('Merge files "input_file1.pdf" and "input_file2.pdf" into "output_file.pdf"')}*:

- `file_conversor pdf merge "input_file1.pdf" "input_file2.pdf" -o output_file.pdf` 



*{_('Merge protected PDF "input_file1.pdf" with password "unlock_password" with unprotected file "input_file2.pdf"')}*:

- `file_conversor pdf merge "input_file1.pdf:unlock_password" "input_file2.pdf" -o output_file.pdf` 
    """)
def merge(
    input_files: Annotated[List[str], typer.Argument(help=f"{_('Input file')} ({', '.join(PyPDFBackend.SUPPORTED_IN_FORMATS)}). {_('If file is protected, provide its password using the format `"filepath:password"`')}.",
                                                     )],
    output_file: Annotated[str | None, typer.Option("--output", "-o",
                                                    help=f"{_("Output file")} ({', '.join(PyPDFBackend.SUPPORTED_OUT_FORMATS)}). {_('Defaults to None')} ({_('use the same input file of 1st file as output name')}, {_('with _merged.pdf at the end)')}",
                                                    callback=lambda x: check_file_format(x, PyPDFBackend.SUPPORTED_OUT_FORMATS),
                                                    )] = None,
):
    pypdf_backend = PyPDFBackend(verbose=STATE["verbose"])
    with get_progress_bar() as progress:
        merge_task = progress.add_task(f"{_('Processing file')} '{output_file}':", total=None,)

        # get dict in format {filepath: password}
        filepath_dict = {}
        FILEPATH_RE = re.compile(r'^(.+?)(::(.+))?$')
        for arg in input_files:
            match = FILEPATH_RE.search(arg)
            if not match:
                raise RuntimeError(f"{_('Invalid filepath format')} '{arg}'. {_("Valid format is 'filepath:password' or 'filepath'")}.")

            # check user input
            filepath = match.group(1)
            password = match.group(3) if match.group(3) else None

            # create filepath_dict
            filepath_dict[filepath] = password

        pypdf_backend.merge(
            # files
            input_files=filepath_dict,
            output_file=output_file if output_file else f"{input_files[0].replace(".pdf", "")}_merged.pdf",
        )
        progress.update(merge_task, total=100, completed=100)

    logger.info(f"{_('Merge pages')}: [bold green]{_('SUCCESS')}[/].")


# pdf split
@pdf_cmd.command(
    help=f"""
        {_('Split PDF pages into several 1-page PDFs.')}

        {_('For every PDF page, a new single page PDF will be created using the format `output_file_X.pdf`, where X is the page number.')}
    """,
    epilog=f"""
**{_('Examples')}:** 



*{_('Split pages of input_file.pdf into output_file_X.pdf files')}*:

- `file_conversor pdf split input_file.pdf -o output_file.pdf` 



*{_('For every PDF page, generate a "input_file_X.pdf" file')}*:

- `file_conversor pdf split input_file.pdf` 
""")
def split(
    input_file: Annotated[str, typer.Argument(help=f"{_('Input file')} ({', '.join(PyPDFBackend.SUPPORTED_IN_FORMATS)})",
                                              callback=lambda x: check_file_format(x, PyPDFBackend.SUPPORTED_IN_FORMATS, exists=True),
                                              )],
    output_file: Annotated[str | None, typer.Option("--output", "-o",
                                                    help=f"{_('Output file path')} ({', '.join(PyPDFBackend.SUPPORTED_OUT_FORMATS)}). {_('Defaults to None')} ({_('use the same input file as output name')})",
                                                    callback=lambda x: check_file_format(x, PyPDFBackend.SUPPORTED_OUT_FORMATS),
                                                    )] = None,
    decrypt_password: Annotated[str | None, typer.Option("--password", "-p",
                                                         help=_("Password used to open protected file. Defaults to None (do not decrypt)."),
                                                         )] = None,
):
    pypdf_backend = PyPDFBackend(verbose=STATE["verbose"])
    with get_progress_bar() as progress:
        split_task = progress.add_task(f"{_('Processing file')} '{input_file}':", total=None,)

        pypdf_backend.split(
            # files
            input_file=input_file,
            output_file=output_file if output_file else input_file,

            # passwords
            decrypt_password=decrypt_password,
        )
        progress.update(split_task, total=100, completed=100)

    logger.info(f"{_('Split pages')}: [bold green]{_('SUCCESS')}[/].")


# pdf extract
@pdf_cmd.command(
    help=f"""
        {_('Extract specific pages from a PDF.')}
    """,
    epilog=f"""
**{_('Examples')}:** 



*{_('Extract pages 1 to 2, 4 and 6')}*:

- `file_conversor pdf extract input_file.pdf -o output_file.pdf -pg 1-2 -pg 4 -pg 6` 
    """)
def extract(
    input_file: Annotated[str, typer.Argument(help=f"{_('Input file')} ({', '.join(PyPDFBackend.SUPPORTED_IN_FORMATS)})",
                                              callback=lambda x: check_file_format(x, PyPDFBackend.SUPPORTED_IN_FORMATS, exists=True),
                                              )],
    output_file: Annotated[str | None, typer.Option("--output", "-o",
                                                    help=f"{_('Output file')} ({', '.join(PyPDFBackend.SUPPORTED_OUT_FORMATS)}) {_('Defaults to None')} ({_('use the same input file as output name')}, {_('with _extracted.pdf at the end)')}",
                                                    callback=lambda x: check_file_format(x, PyPDFBackend.SUPPORTED_OUT_FORMATS),
                                                    )] = None,
    pages: Annotated[List[str] | None, typer.Option("--pages", "-pg",
                                                    help=_('Pages to extract (comma-separated list). Format "page_num" or "start-end".'),
                                                    )] = None,
    decrypt_password: Annotated[str | None, typer.Option("--password", "-p",
                                                         help=_("Password used to open protected file. Defaults to None (do not decrypt)."),
                                                         )] = None,
):
    if not pages:
        pages_str = typer.prompt(f"{_('Pages to extract [comma-separated list] (e.g., 1-3, 7)')}")
        pages = [p.strip() for p in str(pages_str).split(",")]

    pypdf_backend = PyPDFBackend(verbose=STATE["verbose"])
    with get_progress_bar() as progress:
        extract_task = progress.add_task(f"{_('Processing file')} '{input_file}':", total=None,)

        # parse user input
        pages_list = []
        PAGES_RE = re.compile(r'(\d+)(-(\d*)){0,1}')
        for arg in pages:
            match = PAGES_RE.search(arg)
            if not match:
                raise RuntimeError(f"{_('Invalid page instruction')} '{arg}'. {_("Valid format is 'begin-', 'begin-end', 'page_num'")}.")

            # check user input
            begin = int(match.group(1)) - 1
            end = begin
            if match.group(3):
                end = int(match.group(3)) - 1
            elif match.group(2):
                end = pypdf_backend.len(input_file)

            if end < begin:
                raise RuntimeError(f"{_('Invalid begin-end page interval')}. {_('End Page < Begin Page')} '{arg}'.")

            # create pages list
            for page_num in range(begin, end + 1):
                pages_list.append(page_num)

        pypdf_backend.extract(
            # files
            input_file=input_file,
            output_file=output_file if output_file else f"{input_file.replace(".pdf", "")}_extracted.pdf",

            # passwords
            decrypt_password=decrypt_password,

            # other args
            pages=pages_list
        )
        progress.update(extract_task, total=100, completed=100)

    logger.info(f"{_('Extract pages')}: [bold green]{_('SUCCESS')}[/].")


# pdf rotate
@pdf_cmd.command(
    help=f"""
        {_('Rotate PDF pages (clockwise or anti-clockwise).')}
    """,
    epilog=f"""
**{_('Examples')}:** 



*{_('Rotate page 1 by 180 degress')}*:

- `file_conversor pdf rotate input_file.pdf -o output_file.pdf -r "1:180"` 



*{_('Rotate page 5-7 by 90 degress, 9 by -90 degrees, 10-15 by 180 degrees')}*:

- `file_conversor pdf rotate input_file.pdf -o output_file.pdf -r "5-7:90" -r "9:-90" -r "10-15:180"`
    """)
def rotate(
    input_file: Annotated[str, typer.Argument(help=f"{_('Input file')} ({', '.join(PyPDFBackend.SUPPORTED_IN_FORMATS)})",
                                              callback=lambda x: check_file_format(x, PyPDFBackend.SUPPORTED_IN_FORMATS, exists=True),
                                              )],
    rotation: Annotated[List[str], typer.Option("--rotation", "-r",
                                                help=_("List of pages to rotate. Format ``\"page:rotation\"`` or ``\"start-end:rotation\"`` or ``\"start-:rotation\"`` ..."),
                                                )],
    output_file: Annotated[str | None, typer.Option("--output", "-o",
                                                    help=f"{_('Output file')} ({', '.join(PyPDFBackend.SUPPORTED_OUT_FORMATS)}). {_('Defaults to None')} ({_('use the same input file as output name')}, {_('with _rotated.pdf at the end)')}",
                                                    callback=lambda x: check_file_format(x, PyPDFBackend.SUPPORTED_OUT_FORMATS),
                                                    )] = None,
    decrypt_password: Annotated[str | None, typer.Option("--password", "-p",
                                                         help=_("Password used to open protected file. Defaults to None (do not decrypt)."),
                                                         )] = None,
):
    pypdf_backend = PyPDFBackend(verbose=STATE["verbose"])
    with get_progress_bar() as progress:
        rotate_task = progress.add_task(f"{_('Processing file')} '{input_file}':", total=None,)

        # get rotation dict in format {page: rotation}
        rotation_dict = {}
        ROTATION_RE = re.compile(r'(\d+)(-(\d*)){0,1}:([-]{0,1}\d+)')
        for arg in rotation:
            match = ROTATION_RE.search(arg)
            if not match:
                raise RuntimeError(f"{_('Invalid rotation instruction')} '{arg}'. {_("Valid format is 'begin-end:degree' or 'page:degree'")}.")

            # check user input
            begin = int(match.group(1)) - 1
            end = begin
            if match.group(3):
                end = int(match.group(3)) - 1
            elif match.group(2):
                end = pypdf_backend.len(input_file)
            degree = int(match.group(4))
            if end < begin:
                raise RuntimeError(f"{_('Invalid begin-end page interval')}. {_('End Page < Begin Page')} '{arg}'.")

            # create rotation_dict
            for page_num in range(begin, end + 1):
                rotation_dict[page_num] = degree

        pypdf_backend.rotate(
            # files
            input_file=input_file,
            output_file=output_file if output_file else f"{input_file.replace(".pdf", "")}_rotated.pdf",

            # passwords
            decrypt_password=decrypt_password,

            # other args
            rotations=rotation_dict,
        )
        progress.update(rotate_task, total=100, completed=100)

    logger.info(f"{_('Rotate pages')}: [bold green]{_('SUCCESS')}[/].")


# pdf encrypt
@pdf_cmd.command(
    rich_help_panel=SECURITY_PANEL,
    help=f"""
        {_('Protect PDF file with a password (create encrypted PDF file).')}
    """,
    epilog=f"""
        **{_('Examples')}:** 

        - `file_conversor pdf encrypt input_file.pdf -o output_file.pdf --owner-password 1234`

        - `file_conversor pdf encrypt input_file.pdf -o output_file.pdf -op 1234 --up 0000 -an -co`
    """)
def encrypt(
    input_file: Annotated[str, typer.Argument(help=f"{_('Input file')} ({', '.join(PyPDFBackend.SUPPORTED_IN_FORMATS)})",
                                              callback=lambda x: check_file_format(x, PyPDFBackend.SUPPORTED_IN_FORMATS, exists=True),
                                              )],
    owner_password: Annotated[str, typer.Option("--owner-password", "-op",
                                                help=_("Owner password for encryption. Owner has ALL PERMISSIONS in the output PDF file."),
                                                prompt=f"{_('Owner password for encryption (password will not be displayed, for your safety)')}",
                                                hide_input=True,
                                                )],

    output_file: Annotated[str | None, typer.Option("--output", "-o",
                                                    help=f"{_('Output file')} ({', '.join(PyPDFBackend.SUPPORTED_OUT_FORMATS)}). {_('Defaults to None')} ({_('use the same input file as output name')}, {_('with _encrypted.pdf at the end)')}",
                                                    callback=lambda x: check_file_format(x, PyPDFBackend.SUPPORTED_OUT_FORMATS),
                                                    )] = None,
    user_password: Annotated[str | None, typer.Option("--user-password", "-up",
                                                      help=_("User password for encryption. User has ONLY THE PERMISSIONS specified in the arguments. Defaults to None (user and owner password are the same)."),
                                                      )] = None,
    decrypt_password: Annotated[str | None, typer.Option("--decrypt-password", "-dp",
                                                         help=_("Decrypt password used to open protected file. Defaults to None (do not decrypt)."),
                                                         )] = None,

    allow_annotate: Annotated[bool, typer.Option("--annotate", "-an",
                                                 help=_("User can add/modify annotations (comments, highlight text, etc) and interactive forms. Default to False (not set)."),
                                                 is_flag=True,
                                                 )] = False,
    allow_fill_forms: Annotated[bool, typer.Option("--fill-forms", "-ff",
                                                   help=_("User can fill form fields (subset permission of --annotate). Default to False (not set)."),
                                                   is_flag=True,
                                                   )] = False,
    allow_modify: Annotated[bool, typer.Option("--modify", "-mo",
                                               help=_("User can modify the document (e.g., add / edit text, add / edit images, etc). Default to False (not set)."),
                                               is_flag=True,
                                               )] = False,
    allow_modify_pages: Annotated[bool, typer.Option("--modify-pages", "-mp",
                                                     help=_("User can insert, delete, or rotate pages (subset of --modify). Default to False (not set)."),
                                                     is_flag=True,
                                                     )] = False,
    allow_copy: Annotated[bool, typer.Option("--copy", "-co",
                                             help=_("User can copy text/images. Default to False (not set)."),
                                             is_flag=True,
                                             )] = False,
    allow_accessibility: Annotated[bool, typer.Option("--accessibility", "-ac",
                                                      help=_("User can use screen readers for accessibility. Default to True (allow)."),
                                                      is_flag=True,
                                                      )] = True,
    allow_print_lq: Annotated[bool, typer.Option("--print-lq", "-pl",
                                                 help=_("User can print (low quality). Defaults to True (allow)."),
                                                 is_flag=True,
                                                 )] = True,
    allow_print_hq: Annotated[bool, typer.Option("--print-hq", "-ph",
                                                 help=_("User can print (high quality). Requires --allow-print. Defaults to True (allow)."),
                                                 is_flag=True,
                                                 )] = True,
    allow_all: Annotated[bool, typer.Option("--all", "-all",
                                            help=_("User has ALL PERMISSIONS. If set, it overrides all other permissions. Defaults to False (not set)."),
                                            is_flag=True,
                                            )] = False,
    encrypt_algo: Annotated[str, typer.Option("--encryption", "-enc",
                                              help=_("Encryption algorithm used. Valid options are RC4-40, RC4-128, AES-128, AES-256-R5, or AES-256. Defaults to AES-256 (for enhanced security and compatibility)."),
                                              callback=lambda x: check_valid_options(x, valid_options=[None, "RC4-40", "RC4-128", "AES-128", "AES-256-R5", "AES-256"])
                                              )] = "AES-256",
):
    pypdf_backend = PyPDFBackend(verbose=STATE["verbose"])
    with get_progress_bar() as progress:
        encrypt_task = progress.add_task(f"{_('Processing file')} '{input_file}':", total=None,)
        pypdf_backend.encrypt(
            # files
            input_file=input_file,
            output_file=output_file if output_file else f"{input_file.replace(".pdf", "")}_encrypted.pdf",

            # passwords
            owner_password=owner_password,
            user_password=user_password,
            decrypt_password=decrypt_password,

            # permissions
            permission_annotate=allow_annotate,
            permission_fill_forms=allow_fill_forms,
            permission_modify=allow_modify,
            permission_modify_pages=allow_modify_pages,
            permission_copy=allow_copy,
            permission_accessibility=allow_accessibility,
            permission_print_low_quality=allow_print_lq,
            permission_print_high_quality=allow_print_hq,
            permission_all=allow_all,
            encryption_algorithm=encrypt_algo,
        )
        progress.update(encrypt_task, total=100, completed=100)

    logger.info(f"{_('Encryption')}: [bold green]{_('SUCCESS')}[/].")


# pdf decrypt
@pdf_cmd.command(
    rich_help_panel=SECURITY_PANEL,
    help=f"""
        {_('Remove password protection from a PDF file  (create decrypted PDF file).')}        
    """,
    epilog=f"""
        **{_('Examples')}:** 

        - `file_conversor pdf decrypt input_file.pdf output_file.pdf --password 1234`

        - `file_conversor pdf decrypt input_file.pdf output_file.pdf -p 1234`
    """)
def decrypt(
    input_file: Annotated[str, typer.Argument(help=f"{_('Input file')} ({', '.join(PyPDFBackend.SUPPORTED_IN_FORMATS)})",
                                              callback=lambda x: check_file_format(x, PyPDFBackend.SUPPORTED_IN_FORMATS, exists=True),
                                              )],
    password: Annotated[str, typer.Option("--password", "-p",
                                          help=_("Password used for decryption."),
                                          prompt=f"{_('Password for decryption (password will not be displayed, for your safety)')}",
                                          hide_input=True,
                                          )],

    output_file: Annotated[str | None, typer.Option("--output", "-o",
                                                    help=f"{_('Output file')} ({', '.join(PyPDFBackend.SUPPORTED_OUT_FORMATS)}). {_('Defaults to None')} ({_('use the same input file as output name')}, {_('with _decrypted.pdf at the end)')}",
                                                    callback=lambda x: check_file_format(x, PyPDFBackend.SUPPORTED_OUT_FORMATS),
                                                    )] = None,
):
    pypdf_backend = PyPDFBackend(verbose=STATE["verbose"])
    with get_progress_bar() as progress:
        decrypt_task = progress.add_task(f"{_('Processing file')} '{input_file}':", total=None,)
        pypdf_backend.decrypt(
            input_file=input_file,
            output_file=output_file if output_file else f"{input_file.replace(".pdf", "")}_decrypted.pdf",
            password=password,
        )
        progress.update(decrypt_task, total=100, completed=100)

    logger.info(f"{_('Decryption')}: [bold green]{_('SUCCESS')}[/].")



# src\file_conversor\utils\rich.py


from types import TracebackType
from typing import Any, Optional, Self, Type

from rich.progress import Progress, TaskID, TextColumn, BarColumn, TimeRemainingColumn

# user-provided
from file_conversor.config.state import State

STATE = State.get_instance()


class DummyProgress:
    def __init__(self, *args, **kwargs) -> None:
        super().__init__()

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        pass

    def add_task(self, description: str, start: bool = True, total: float | None = 100, completed: int = 0, visible: bool = True, **fields: Any) -> TaskID:
        return TaskID(0)

    def update(self, task_id: TaskID, *, total: float | None = None, completed: float | None = None, advance: float | None = None, description: str | None = None, visible: bool | None = None, refresh: bool = False, **fields: Any):
        pass

    def advance(self, task_id: TaskID, advance: float = 1) -> None:
        pass


def get_progress_bar() -> Progress | DummyProgress:
    """Gets rich Progress() instance, properly formatted"""

    if STATE['no-progress']:
        return DummyProgress()
    return Progress(
        TextColumn("[bold blue]{task.description}"),
        BarColumn(bar_width=40),
        "[bold white][progress.percentage]{task.percentage:>3.0f}%",
        TimeRemainingColumn(),
    )




# src\file_conversor\system\win\utils.py

import os
import platform
import subprocess
import time

# Import winreg only on Windows to avoid ImportError on other OSes
if platform.system() == "Windows":
    import winreg
    import ctypes
else:
    # Placeholder so the names exists
    winreg = None
    ctypes = None


def is_admin() -> bool:
    """True if app running with admin priviledges, False otherwise."""
    try:
        if ctypes:
            return ctypes.windll.shell32.IsUserAnAdmin()
    except:
        pass
    return False


def reload_user_path():
    """Reload user PATH in current process."""
    if winreg is None:
        return
    with winreg.OpenKey(winreg.HKEY_CURRENT_USER, r"Environment") as key:
        user_path, _ = winreg.QueryValueEx(key, "PATH")
        os.environ["PATH"] = user_path + os.pathsep + os.environ["PATH"]


def restart_explorer():
    # Step 1: kill explorer.exe
    subprocess.run(
        ["taskkill", "/f", "/im", "explorer.exe"],
        capture_output=True,
        text=True,  # Capture output as text (Python 3.7+)
        check=True,
    )
    # Wait briefly to ensure process termination
    time.sleep(0.5)  # Increased delay for stability
    # Step 2: Restart explorer.exe
    subprocess.Popen(
        "explorer.exe",
        shell=True,
        stdin=subprocess.DEVNULL,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
        close_fds=True,  # Detach from Typer
    )



# src\file_conversor\backend\img2pdf_backend.py

"""
This module provides functionalities for handling PDF files using ``img2pdf`` backend.
"""

import img2pdf

from datetime import datetime
from typing import Any, Iterable

# user-provided imports
from file_conversor.backend.abstract_backend import AbstractBackend


class Img2PDFBackend(AbstractBackend):
    """
    A class that provides an interface for handling PDF files using ``img2pdf``.
    """

    FIT_MODE = img2pdf.FitMode

    FIT_INTO = img2pdf.FitMode.into
    FIT_FILL = img2pdf.FitMode.fill

    LAYOUT_NONE = None  # pdf exact size as image
    LAYOUT_A4_PORTRAIT_CM = (21.00, 29.70)  # A4 in cm
    LAYOUT_A4_LANDSCAPE_CM = (LAYOUT_A4_PORTRAIT_CM[1], LAYOUT_A4_PORTRAIT_CM[0])

    SUPPORTED_IN_FORMATS = {
        "jpeg": {},
        "jpg": {},
        "png": {},
        "tiff": {},
        "tif": {},
        "bmp": {},
        "gif": {},
    }
    SUPPORTED_OUT_FORMATS = {
        "pdf": {},
    }

    def __init__(
        self,
        verbose: bool = False,
    ):
        """
        Initialize the ``pypdf`` backend

        :param verbose: Verbose logging. Defaults to False.      
        """
        super().__init__()
        self._verbose = verbose

    def to_pdf(self,
               output_file: str,
               input_files: Iterable[str],
               image_fit: FIT_MODE = FIT_INTO,
               page_size: tuple[float, float] | None = LAYOUT_NONE,
               dpi: int = 200,
               include_metadata: bool = True
               ):
        """
        Convert input image files into one PDF output file.

        image_fit = 
        - FIT_INTO: resize image to fit in page (keep proportions, DO NOT cut image)
        - FIT_FILL: resize image to page - no borders allowed (keep proportions, CUT image if necessary)

        dpi = 
        -  96 = low quality, for screen.
        - 200 = good quality, for screen.
        - 300 = high quality, for printing.

        :param output_file: Output PDF file
        :param input_files: Input image files. 
        :param image_fit: Where and how to place fig. Valid only when ``page_size`` != None. Defaults to FIT_INTO.
        :param page_size: PDF page size, in centimeters (cm). Format (width, height). Defaults to LAYOUT_NONE (PDF size is exactly the size of figs).
        :param dpi: Set dots per inch (DPI) for picture. Defaults to 200.
        :param include_metadata: Include basic metadata (moddata, createdate, creator, etc). Defaults to True.

        :raises FileNotFoundError: if input file not found
        """
        for input_file in input_files:
            self.check_file_exists(input_file)
        output_file = output_file.replace(".pdf", "")

        # get current day
        now = datetime.now()
        opts: dict[str, Any] = {
            'dpi': dpi,
        }

        # metadata
        if include_metadata:
            opts.update({
                # PDF metadata
                'creationdate': now,
                'moddate': now,
                'creator': "img2pdf",
                'producer': "img2pdf",
            })

        # page layout
        if page_size:
            opts['layout_fun'] = img2pdf.get_layout_fun(
                pagesize=tuple(img2pdf.cm_to_pt(x) for x in page_size),
                fit=image_fit,
            )

        with open(f"{output_file}.pdf", "wb") as f:
            f.write(img2pdf.convert(*input_files, **opts))



# src\file_conversor\utils\validators.py

import typer

from pathlib import Path
from typing import Any, Iterable, List

# user provided imports
from file_conversor.config.locale import get_translation

from file_conversor.utils.file import File

_ = get_translation()


def check_is_bool_or_none(data: str | bool | None) -> bool | None:
    """
    Checks if the provided input is a valid bool or None.
    """
    if data is None or isinstance(data, bool):
        return data
    if isinstance(data, str):
        if data.lower() == "true":
            return True
        if data.lower() == "false":
            return False
        if data.lower() == "none":
            return None
    raise typer.BadParameter(_("Must be a bool or None."))


def check_positive_integer(bitrate: int | float):
    """
    Checks if the provided number is a positive integer.
    """
    if bitrate <= 0:
        raise typer.BadParameter(_("Must be a positive integer."))
    return bitrate


def check_file_format(filename_or_iter: list | dict | set | str | None, format_dict: dict | list, exists: bool = False):
    """
    Checks if the provided format is supported.

    :param filename_or_iter: Filename or iterable list
    :param format_dict: Format {format:options} or [format]
    :param exists: Check if file exists. Default False (do not check).

    :raises typer.BadParameter: Unsupported format, or file not found.
    :raises TypeError: Invalid parameter type.
    """
    file_list = []
    if isinstance(filename_or_iter, (list, dict, set)):
        file_list = list(filename_or_iter)
    elif isinstance(filename_or_iter, str):
        file_list.append(filename_or_iter)
    elif filename_or_iter is None:
        return filename_or_iter
    else:
        raise TypeError(f"{_('Invalid type')} '{type(filename_or_iter)}' {_('for')} filename_or_iter. {_('Valid values are Iterable | str | None')}.")
    for filename in file_list:
        file = File(filename)
        file_format = file.get_extension().lower()
        if file_format not in format_dict:
            raise typer.BadParameter(f"\n{_('Unsupported format')} '{file_format}'. {_('Supported formats are')}: {', '.join(format_dict)}.")
        if exists and not file.is_file():
            raise typer.BadParameter(f"{_("File")} '{filename}' {_("not found")}")
    return filename_or_iter


def check_valid_options(data: Any | None, valid_options: Iterable):
    if data not in valid_options:
        raise typer.BadParameter(f"'{data}' is invalid.  Valid options are {', '.join(valid_options)}.")
    return data




# src\file_conversor\__main__.py

import subprocess
import sys

from rich import print

# user provided imports
from file_conversor.cli.app_cmd import app_cmd, STATE, CONFIG, logger, _
from file_conversor.system import reload_user_path


# Entry point of the app
def main():
    try:
        # begin app
        reload_user_path()
        app_cmd()
    except Exception as e:
        error_type = str(type(e)).split("'")[1]
        logger.error(f"{error_type} ({e})", exc_info=True if STATE["debug"] else None)
        if isinstance(e, subprocess.CalledProcessError):
            logger.error(f"CMD: {e.cmd} ({e.returncode})")
            logger.error(f"STDERR: {e.stderr}")
            logger.error(f"STDOUT: {e.stdout}")
        if STATE["debug"]:
            raise
        sys.exit(1)


# Start the application
if __name__ == "__main__":
    main()



# src\file_conversor\backend\win_reg_backend.py

"""
This module provides functionalities for handling Windows Registry using ``reg`` backend.
"""

import subprocess
import tempfile

from pathlib import Path

# user-provided imports
from file_conversor.config.log import Log

from file_conversor.system.win import WinRegFile
from file_conversor.backend.abstract_backend import AbstractBackend

LOG = Log.get_instance()

logger = LOG.getLogger(__name__)


class WinRegBackend(AbstractBackend):
    """
    A class that provides an interface for handling .REG files using ``reg``.
    """

    SUPPORTED_IN_FORMATS = {
        "reg": {},
    }
    SUPPORTED_OUT_FORMATS = {
        "reg": {},
    }

    def __init__(self, verbose: bool = False):
        """
        Initialize the ``reg`` backend

        :param verbose: Verbose logging. Defaults to False.      
        """
        super().__init__()
        self._verbose = verbose

        self._reg_bin = self.find_in_path("reg")

    def import_file(
        self,
        input_file_or_winreg: str | Path | WinRegFile,
    ):
        """
        Import registry info from input .REG file.

        :param input_file_or_winreg: Input REG file, or WinRegFile.        

        :raises subprocess.CalledProcessError: if reg cannot import registry file
        """
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_file = (Path(temp_dir) / ".out.reg").resolve().__str__()

            winregfile: WinRegFile
            if isinstance(input_file_or_winreg, WinRegFile):
                winregfile = input_file_or_winreg
            else:
                winregfile = WinRegFile(input_file_or_winreg)
            winregfile.dump(temp_file)

            # build command
            command = [str(self._reg_bin), "import", temp_file]

            # Execute command
            logger.info(f"Importing .REG file ...")
            logger.debug(" ".join(command))

            subprocess.run(command,
                           capture_output=True,
                           check=True,
                           )

    def delete_keys(
        self,
        input_file_or_winreg: str | Path | WinRegFile,
    ):
        """
        Loads registry keys from input .REG file, and deletes them from windows registry.

        :param input_file_or_winreg: Input .REG file.        

        :raises subprocess.CalledProcessError: if reg cannot delete registry keys
        """
        winregfile: WinRegFile
        if isinstance(input_file_or_winreg, WinRegFile):
            winregfile = input_file_or_winreg
        else:
            winregfile = WinRegFile(input_file_or_winreg)

        logger.info(f"Deleting reg keys ...")
        for _, key in winregfile.items():
            # Execute command
            try:
                subprocess.run([str(self._reg_bin), "query", key.path],
                               stdin=subprocess.DEVNULL,
                               stdout=subprocess.DEVNULL,
                               stderr=subprocess.DEVNULL,
                               check=True,
                               )
            except subprocess.CalledProcessError:
                # logger.debug(f"SKIP - key '{key.path}' not found ...")
                continue

            subprocess.run([str(self._reg_bin), "delete", key.path, "/f"],
                           capture_output=True,
                           check=True,
                           )

            logger.debug(f"'{key.path}' deleted")



# src\file_conversor\backend\ffmpeg_backend.py

"""
This module provides functionalities for handling audio and video files using FFmpeg.
"""

import json
import shutil
import subprocess
import re

from datetime import timedelta
from typing import Iterable

# user-provided imports
from file_conversor.config import Log
from file_conversor.config.locale import get_translation

from file_conversor.utils import File

from file_conversor.dependency import BrewPackageManager, ScoopPackageManager
from file_conversor.backend.abstract_backend import AbstractBackend

_ = get_translation()
LOG = Log.get_instance()

logger = LOG.getLogger(__name__)


class FFmpegBackend(AbstractBackend):
    """
    FFmpegBackend is a class that provides an interface for handling audio and video files using FFmpeg.
    """

    SUPPORTED_IN_AUDIO_FORMATS = {
        'aac': {},
        'ac3': {},
        'flac': {},
        'm4a': {},
        'mp3': {},
        'ogg': {},
        'opus': {},
        'wav': {},
    }
    SUPPORTED_IN_VIDEO_FORMATS = {
        '3gp': {},
        'asf': {},
        'avi': {},
        'flv': {},
        'h264': {},
        'hevc': {},
        'm4v': {},
        'mkv': {},
        'mov': {},
        'mp4': {},
        'mpeg': {},
        'mpg': {},
        'webm': {},
    }
    SUPPORTED_IN_FORMATS = SUPPORTED_IN_AUDIO_FORMATS | SUPPORTED_IN_VIDEO_FORMATS

    SUPPORTED_OUT_AUDIO_FORMATS = {
        'mp3': {
            '-f': 'mp3',
            '-c:a': 'libmp3lame',
        },
        'm4a': {
            '-f': 'ipod',
            '-c:a': 'aac',
        },
        'ogg': {
            '-f': 'ogg',
            '-c:a': 'libvorbis',
        },
        'opus': {
            '-f': 'opus',
            '-c:a': 'libopus',
        },
        'flac': {
            '-f': 'flac',
            '-c:a': 'flac',
        },
    }
    SUPPORTED_OUT_VIDEO_FORMATS = {
        'mp4': {
            '-f': 'mp4',
            '-c:v': 'libx264',
            '-c:a': 'aac',
        },
        'avi': {
            '-f': 'avi',
            '-c:v': 'mpeg4',
            '-c:a': 'libmp3lame',
        },
        'mkv': {
            '-f': 'matroska',
            '-c:v': 'libx264',
            '-c:a': 'aac',
        },
        'webm': {
            '-f': 'webm',
            '-c:v': 'libvpx',
            '-c:a': 'libvorbis',
        },
    }
    SUPPORTED_OUT_FORMATS = SUPPORTED_OUT_VIDEO_FORMATS | SUPPORTED_OUT_AUDIO_FORMATS

    PROGRESS_RE = re.compile(r'time=(\d+):(\d+):([\d\.]+)')

    @staticmethod
    def get_convert_progress(process: subprocess.Popen, file_duration_secs: float) -> float:
        """
        Gets FFMpeg current progress status, in percentage (0-100)

        :param process: Subprocess.Popen object.
        :param file_duration_secs: Input file duration (in secs).

        :return: Progress of FFMpeg [0-100], otherwise 0
        """
        if process and process.stderr:
            match = FFmpegBackend.PROGRESS_RE.search(process.stderr.readline())
            if match:
                hours = int(match.group(1))
                minutes = int(match.group(2))
                seconds = float(match.group(3))
                current_time = hours * 3600 + minutes * 60 + seconds
                return 100.0 * (float(current_time) / file_duration_secs)
        return 0

    def __init__(
        self,
        install_deps: bool | None,
    ):
        """
        Initialize the FFMpeg backend.

        :param install_deps: Install external dependencies. If True auto install using a package manager. If False, do not install external dependencies. If None, asks user for action. 

        :raises RuntimeError: if ffmpeg dependency is not found
        """
        super().__init__(
            pkg_managers={
                ScoopPackageManager({
                    "ffmpeg": "ffmpeg"
                }),
                BrewPackageManager({
                    "ffmpeg": "ffmpeg"
                }),
            },
            install_answer=install_deps,
        )

        # check ffprobe / ffmpeg
        self._ffprobe_bin = self.find_in_path("ffprobe")
        self._ffmpeg_bin = self.find_in_path("ffmpeg")

    def calculate_file_total_duration(self, file_path: str) -> float:
        """
        Calculate file total duration (in secs), using `ffprobe`.

        :return: Total duration in seconds.
        """
        result = subprocess.run(
            [str(self._ffprobe_bin), '-v', 'error', '-show_entries',
             'format=duration', '-of',
             'default=noprint_wrappers=1:nokey=1', file_path],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True
        )
        duration_str = result.stdout.strip()
        return float(duration_str if duration_str else "0")

    def calculate_file_formatted_duration(self, file_path: str) -> str:
        """
        Calculate file duration (formatted), using `ffprobe`.

        :return: Total duration  (format HH:MM:SS).
        """
        duration_secs = self.calculate_file_total_duration(file_path)

        # Converte segundos para timedelta e formata como HH:MM:SS
        td = timedelta(seconds=int(duration_secs))
        return str(td)

    def get_file_info(self, file_path: str) -> dict:
        """
        Executa ffprobe e retorna os metadados no formato JSON

        result = {
            streams: [],
            chapters: [],
            format: {},
        }

        stream = {
            index,
            codec_name,
            codec_long_name,
            codec_type: audio|video,
            sampling_rate,
            channels,
            channel_layout: stereo|mono,
        }

        format = {
            format_name,
            format_long_name,
            duration,
            size,
        }

        :return: JSON object
        """
        command = [
            str(self._ffprobe_bin),
            "-v", "quiet",
            "-print_format", "json",
            "-show_format",
            "-show_streams",
            "-show_chapters",
            "-show_error",
            file_path
        ]
        result = subprocess.run(
            command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
        return json.loads(result.stdout)

    def _set_input(self, input_file: str) -> tuple[str, list]:
        """
        Set the input file and check if it has a supported format.

        :param input_file: Input file path.

        :return: (Input file, in options).

        :raises FileNotFoundError: If the input file does not exist.
        :raises ValueError: If the input file format is not supported.
        """
        # check file is found
        in_file = File(input_file)
        if not in_file.is_file():
            raise FileNotFoundError(f"Input file '{input_file}' not found")

        # check if the input file has a supported format
        in_file.check_supported_format(self.SUPPORTED_IN_FORMATS)

        # set the input format options based on the file extension
        in_opts = []
        in_ext = in_file.get_extension()
        for opt, value in self.SUPPORTED_IN_FORMATS[in_ext].items():
            in_opts.extend([opt, value])

        return input_file, in_opts

    def _set_output(self, output_file: str) -> tuple[str, list]:
        """
        Set the output file and check if it has a supported format.

        :param output_file: Output file path.

        :return: (Output file, out options).

        :raises FileExistsError: If the output path is a file.
        :raises FileNotFoundError: If the output directory could not be created.
        :raises ValueError: If the output file format is not supported.        
        """
        out_file = File(output_file)

        # create out dir (if it does not exists)
        out_dir = File(out_file.get_full_dirname())
        out_dir.mkdir()

        # check if the output file has a supported format
        File(output_file).check_supported_format(self.SUPPORTED_OUT_FORMATS)

        # set the output format options based on the file extension
        out_opts = []
        out_ext = out_file.get_extension()
        for opt, value in self.SUPPORTED_OUT_FORMATS[out_ext].items():
            out_opts.extend([opt, value])

        return output_file, out_opts

    def convert(
        self,
            input_file: str,
            output_file: str,
            overwrite_output: bool = True,
            stats: bool = False,
        verbose: bool = False,
            in_options: Iterable | None = None,
            out_options: Iterable | None = None,
    ) -> subprocess.Popen:
        """
        Execute the FFmpeg command to convert the input file to the output file.

        :param input_file: Input file path.
        :param output_file: Output file path.      
        :param overwrite_output: Overwrite output file (no user confirmation prompt). Defaults to True.      
        :param stats: Show progress stats. Defaults to False.      
        :param verbose: Verbose logging. Defaults to False.      
        :param in_options: Additional input options. Defaults to None.      
        :param out_options: Additional output options. Defaults to None.    

        :return: Subprocess.Popen object

        :raises RuntimeError: If FFmpeg encounters an error during execution.
        """
        # set input/output files and options
        in_file, in_opts = self._set_input(input_file)
        out_file, out_opts = self._set_output(output_file)

        in_opts.extend(in_options if in_options else [])
        out_opts.extend(out_options if out_options else [])

        # set global options
        global_options = [
            # overwrite output (no confirm)
            "-y" if overwrite_output else "-n",
            "-v" if verbose else "",  # verbose output
            "-stats" if stats else "",  # print progress stats
        ]

        # build ffmpeg command
        ffmpeg_command = []
        ffmpeg_command.extend([str(self._ffmpeg_bin)])  # ffmpeg CLI
        ffmpeg_command.extend(global_options)    # set global options
        ffmpeg_command.extend(in_opts)           # set in options
        ffmpeg_command.extend(["-i", in_file])   # set input
        ffmpeg_command.extend(out_opts)          # set out options
        ffmpeg_command.extend([out_file])        # set output

        # remove empty strings
        ffmpeg_command = [arg for arg in ffmpeg_command if arg != ""]

        logger.info(f"Executing FFmpeg ...")
        logger.debug(f"{" ".join(ffmpeg_command)}")

        # Execute the FFmpeg command
        _convert_process = subprocess.Popen(
            ffmpeg_command,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True
        )
        return _convert_process
